
/*
//  GeMSS128-ring
//  Created by Viliam Hromada on 3/31/21
//  Original by Bassham, Lawrence E (Fed) on 8/29/17.
//  Copyright © 2017 Bassham, Lawrence E (Fed). All rights reserved 
//  This implementation of Ring GeMSS128 is a modification of the official GeMSS reference implementation
//  available at NIST PQC Round 3, https://csrc.nist.gov/CSRC/media/Projects/post-quantum-cryptography/documents/round-3/submissions/GeMSS-Round3.zip
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "rng.h"
#include "api.h"
/* XXX I add this file for randombytes XXX */
#include "randombytes.h"

#include "sign_openHFE.h"
#include "signHFE.h"

#define NB_KATS 1
#define	MAX_MARKER_LEN		50

#define KAT_SUCCESS          0
#define KAT_FILE_OPEN_ERROR -1
#define KAT_DATA_ERROR      -3
#define KAT_CRYPTO_FAILURE  -4

#define NUMBER_OF_USERS 5
#define NUMBER_OF_EXPERIMENTS 100

int		FindMarker(FILE *infile, const char *marker);
int		ReadHex(FILE *infile, unsigned char *A, int Length, char *str);
void	fprintBstr(FILE *fp, char *S, unsigned char *A, unsigned long long L);

#ifdef GeMSS
    char    AlgName[] = "GeMSS";
#elif defined(BlueGeMSS)
    char    AlgName[] = "BlueGeMSS";
#elif defined(RedGeMSS)
    char    AlgName[] = "RedGeMSS";
#elif defined(WhiteGeMSS)
    char    AlgName[] = "WhiteGeMSS";
#elif defined(CyanGeMSS)
    char    AlgName[] = "CyanGeMSS";
#elif defined(MagentaGeMSS)
    char    AlgName[] = "MagentaGeMSS";
#elif defined(FGeMSS)
    char    AlgName[] = "FGeMSS";
#elif defined(DualModeMS)
    char    AlgName[] = "DualModeMS";
#else
    char    AlgName[] = "MQsoft";
#endif

int
main()
{
    char                fn_req[32], fn_rsp[32], fn_int[32];
    FILE                *fp_req, *fp_rsp, *fp_int;
    unsigned char       seed[48];
    unsigned char       entropy_input[48];
    unsigned char       *m, *sm, *m1;
    unsigned long long  mlen, smlen, mlen1;
    int                 count;
    int                 done;
    FILE* output_sig, *output_ver;
    /* XXX Modification to use malloc for large sizes of keys XXX */
    /*  unsigned char       pk[CRYPTO_PUBLICKEYBYTES], sk[CRYPTO_SECRETKEYBYTES]; */
    unsigned char       *pk,*sk;
    int                 ret_val;
    int                 i;
    clock_t 			clock1, clock2;
    clock_t				clock_sign_sum, clock_verification_sum;
    struct timespec		begin,end;
    char fileName[80];
    char fileName2[80];
    sprintf(fileName,"GeMSS_K=%d_NOU=%d_n=%d_delta=%d_v=%d_D=%d_sig",K,NUMBER_OF_USERS,HFEn,HFEDELTA,HFEv,HFEDeg);
    sprintf(fileName2,"GeMSS_K=%d_NOU=%d_n=%d_delta=%d_v=%d_D=%d_ver",K,NUMBER_OF_USERS,HFEn,HFEDELTA,HFEv,HFEDeg);
    output_sig = fopen(fileName,"w");
    output_ver = fopen(fileName2,"w");
    fprintf(output_sig, "%s\n", CRYPTO_ALGNAME );
    fprintf(output_sig, "N: %d delta: %d v: %d D:%d\n",HFEn, HFEDELTA, HFEv,HFEDeg);
    fprintf(output_sig,"users: %d\n",NUMBER_OF_USERS);
    fprintf(output_sig,"pk size [B]: %lu\n",  NUMBER_OF_USERS*CRYPTO_PUBLICKEYBYTES );
    fprintf(output_sig,"signature size [b]: %d\n\n", NUMBER_OF_USERS*CRYPTO_BYTES*8 );

    fprintf(output_ver, "%s\n", CRYPTO_ALGNAME );
    fprintf(output_ver, "N: %d delta: %d v: %d D:%d\n",HFEn, HFEDELTA, HFEv,HFEDeg);
    fprintf(output_ver,"users: %d\n",NUMBER_OF_USERS);
    fprintf(output_ver,"pk size [B]: %lu\n",  NUMBER_OF_USERS*CRYPTO_PUBLICKEYBYTES );
    fprintf(output_ver,"signature size [b]: %d\n\n", NUMBER_OF_USERS*CRYPTO_BYTES*8 );

	
    srand(time(NULL));
    for (i=0; i<48; i++)
		entropy_input[i] = rand()%256;

    randombytes_init(entropy_input, NULL, 256);
    
    randombytes(seed, 48);
    fprintBstr(stderr, (char*)"seed = ", seed, 48);
    mlen = 32;
    fprintf(output_sig, "mlen = %llu\n", mlen);
    fprintf(output_sig, "pk = %llu\n",CRYPTO_PUBLICKEYBYTES*sizeof(unsigned char));
    fprintf(output_sig, "sk = %llu\n",CRYPTO_SECRETKEYBYTES*sizeof(unsigned char));
    fprintf(output_sig, "smlen = %llu\n",mlen+CRYPTO_BYTES);
    fprintf(output_sig, "s = %llu\n",CRYPTO_BYTES);

    fprintf(output_ver, "mlen = %llu\n", mlen);
    fprintf(output_ver, "pk = %llu\n",CRYPTO_PUBLICKEYBYTES*sizeof(unsigned char));
    fprintf(output_ver, "sk = %llu\n",CRYPTO_SECRETKEYBYTES*sizeof(unsigned char));
    fprintf(output_ver, "smlen = %llu\n",mlen+CRYPTO_BYTES);
    fprintf(output_ver, "s = %llu\n",CRYPTO_BYTES);
    fflush(output_ver); fflush(output_sig);
    clock_sign_sum = 0;
    clock_verification_sum = 0;
    fprintf(output_ver,"------------------------\n"); fprintf(output_sig,"------------------------\n");
    for (int number_of_exps = 0; number_of_exps < NUMBER_OF_EXPERIMENTS; number_of_exps++) {
    	    //all pks, sks in one memblock
	    pk=(unsigned char*)malloc(CRYPTO_PUBLICKEYBYTES*sizeof(unsigned char)*NUMBER_OF_USERS);
	    sk=(unsigned char*)malloc(CRYPTO_SECRETKEYBYTES*sizeof(unsigned char)*NUMBER_OF_USERS);
	    
	    randombytes_init(seed, NULL, 256);
	    m = (unsigned char *)calloc(mlen, sizeof(unsigned char));
	    m1 = (unsigned char *)calloc(mlen+CRYPTO_BYTES, sizeof(unsigned char));
	    sm = (unsigned char *)calloc(mlen+CRYPTO_BYTES, sizeof(unsigned char));
		
	    for (i=0; i<48; i++)
		entropy_input[i] = rand()%256;
		
	    randombytes_init(entropy_input, NULL, 256);
		
	    randombytes(m, mlen);
		
	    /* Generate the public/private keypair for each user*/
	    for (int i = 0; i < NUMBER_OF_USERS; i++)
	    {
		if ( (ret_val = crypto_sign_keypair(pk + i*CRYPTO_PUBLICKEYBYTES*sizeof(unsigned char), sk + i*CRYPTO_SECRETKEYBYTES*sizeof(unsigned char))) != 0) {
		    printf("crypto_sign_keypair returned <%d>\n", ret_val);
		    return KAT_CRYPTO_FAILURE;
		}
	    }
    
            /*Ring signature generation - START*/
	    /*Let user u1 be the signature generator*/
	    /*Generate random z2, z3, ..., zk*/
	    //z_vectors will be the result, i.e. z1 = signature, z2 ... zk remain
	    uint8_t *z_vectors = (uint8_t*)malloc( NUMBER_OF_USERS*SIZE_SIGN );
	    UINT* P_z_vectors;
	    P_z_vectors = (UINT*)malloc( NUMBER_OF_USERS*sizeof(UINT)*NB_WORD_GF2nv);
	    memset(P_z_vectors, 0x00, NUMBER_OF_USERS*sizeof(UINT)*NB_WORD_GF2nv);
	    UINT sum_Pi_zi[NB_WORD_GF2nv];
	    memset(sum_Pi_zi, 0x00, sizeof(UINT)*NB_WORD_GF2nv);

	    clock1 = clock();
	    randombytes(z_vectors, NUMBER_OF_USERS*SIZE_SIGN);
	    //evaluate public keys in GeMSS in an interated manner (Feistel-Patarin construction)
	    for (int i = 1; i < NUMBER_OF_USERS; i++)
	    {
	    	sign_openHFE_ring(P_z_vectors + i*NB_WORD_GF2nv, m, mlen, (unsigned char*)(z_vectors + i*SIZE_SIGN), (uint8_t*)(pk + i*CRYPTO_PUBLICKEYBYTES*sizeof(unsigned char)) );
	    }
	    //sum of Pi(zi) of other users
	    for (int i = 1; i < NUMBER_OF_USERS; i++)
	    {
	    	add2_gf2m(sum_Pi_zi,P_z_vectors + i*NB_WORD_GF2nv);
	    }
		

	    //variable sum_Pi_zi contains the sum Pi(zi). This variable will be a parameter of the ring signature of GeMSS
	    //since it is used to initialize the special value S_0
	    //also, the message to be signed is another parameter
	    signHFE_ring(z_vectors,sum_Pi_zi,m,(size_t)mlen,(UINT*)sk);
	    /*variable z_vectors contains the resulting signature*/
	    clock2 = clock();
	    fprintf(output_sig,"%lf\n", (double)(clock2-clock1)/((double)(CLOCKS_PER_SEC)));
    	    clock_sign_sum += clock2-clock1;
	    /*Ring signature generation - END*/
		
	    fflush(output_sig);
	    /*Ring signature verification - START*/
	    UINT w_hat[NB_WORD_GF2nv];
	    UINT digest[SIZE_DIGEST_UINT];
	    UINT cc = 0;
	    memset(w_hat, 0x00, sizeof(UINT)*NB_WORD_GF2nv);
	    clock1 = clock();
	    //evaluate ALL public keys Pi(zi) for all i = 1 ... k
	    for (int i = 0; i < NUMBER_OF_USERS; i++)
	    {
	    	sign_openHFE_ring(P_z_vectors + i*NB_WORD_GF2nv, m, mlen, (unsigned char*)(z_vectors + i*SIZE_SIGN), (uint8_t*)(pk + i*CRYPTO_PUBLICKEYBYTES*sizeof(unsigned char)) );
	    }

	    //sum all the results into w_hat
	    for (int i = 0; i < NUMBER_OF_USERS; i++)
	    {
	    	add2_gf2m(w_hat,P_z_vectors + i*NB_WORD_GF2nv);
	    }
	    //hash the document and normalize according to parameter m
		
	    HASH((unsigned char*)digest,m,mlen);
	    #if HFEmr
	        digest[NB_WORD_GF2m-1]&=MASK_GF2m;
	    #endif
			
	    for(unsigned i=0;i<NB_WORD_GF2m;i++) {
		cc |= (w_hat[i]^digest[i]);
	    }
	    clock2 = clock();
	    if (cc == 0)
	    {
		//fprintf(stderr,"Verification successful!\n");
	    }
	    else
	    {
		//fprintf(stderr,"Verification FAILED!\n");	
	    }	
	    clock_verification_sum += clock2-clock1;
	    fprintf(output_ver,"%lf\n", (double)(clock2-clock1)/((double)(CLOCKS_PER_SEC)));
	    fflush(output_ver); 
	    /*Ring signature verification - END*/

        
	    free(m);
	    free(m1);
	    free(sm);

	    free(pk);
	    free(sk);
    }
    fprintf(output_ver,"------------------------\n"); fprintf(output_sig,"------------------------\n");
    fprintf(output_sig,"Avg time sign: %lf\n", (double)clock_sign_sum/((double)(NUMBER_OF_EXPERIMENTS*CLOCKS_PER_SEC)));
    fprintf(output_ver,"Avg time verify: %lf\n", (double)clock_verification_sum/((double)(NUMBER_OF_EXPERIMENTS*CLOCKS_PER_SEC)));

    fclose(output_sig); fclose(output_ver);
    return KAT_SUCCESS;
}

/*
// ALLOW TO READ HEXADECIMAL ENTRY (KEYS, DATA, TEXT, etc.)
*/
int
FindMarker(FILE *infile, const char *marker)
{
	char	line[MAX_MARKER_LEN];
	int		i, len;
	int curr_line;

	len = (int)strlen(marker);
	if ( len > MAX_MARKER_LEN-1 )
		len = MAX_MARKER_LEN-1;

	for ( i=0; i<len; i++ )
	  {
	    curr_line = fgetc(infile);
	    line[i] = curr_line;
	    if (curr_line == EOF )
	      return 0;
	  }
	line[len] = '\0';

	while ( 1 ) {
		if ( !strncmp(line, marker, len) )
			return 1;

		for ( i=0; i<len-1; i++ )
			line[i] = line[i+1];
		curr_line = fgetc(infile);
		line[len-1] = curr_line;
		if (curr_line == EOF )
		    return 0;
		line[len] = '\0';
	}

	/* shouldn't get here */
	return 0;
}

/*
// ALLOW TO READ HEXADECIMAL ENTRY (KEYS, DATA, TEXT, etc.)
*/
int
ReadHex(FILE *infile, unsigned char *A, int Length, char *str)
{
	int			i, ch, started;
	unsigned char	ich;

	if ( Length == 0 ) {
		A[0] = 0x00;
		return 1;
	}
	memset(A, 0x00, Length);
	started = 0;
	if ( FindMarker(infile, str) )
		while ( (ch = fgetc(infile)) != EOF ) {
			if ( !isxdigit(ch) ) {
				if ( !started ) {
					if ( ch == '\n' )
						break;
					else
						continue;
				}
				else
					break;
			}
			started = 1;
			if ( (ch >= '0') && (ch <= '9') )
				ich = ch - '0';
			else if ( (ch >= 'A') && (ch <= 'F') )
				ich = ch - 'A' + 10;
			else if ( (ch >= 'a') && (ch <= 'f') )
				ich = ch - 'a' + 10;
            else /* shouldn't ever get here */
                ich = 0;
			
			for ( i=0; i<Length-1; i++ )
				A[i] = (A[i] << 4) | (A[i+1] >> 4);
			A[Length-1] = (A[Length-1] << 4) | ich;
		}
	else
		return 0;

	return 1;
}

void
fprintBstr(FILE *fp, char *S, unsigned char *A, unsigned long long L)
{
	unsigned long long  i;

	fprintf(fp, "%s", S);

	for ( i=0; i<L; i++ )
		fprintf(fp, "%02X", A[i]);

	if ( L == 0 )
		fprintf(fp, "00");

	fprintf(fp, "\n");
}

